### Persefone.jl - a model of agricultural landscapes and ecosystems in Europe.
###
### This file holds the code for the Eurasian Skylark (Alauda arvensis).
###

"""
    Skylark

*Alauda arvensis* is a common and charismatic species of agricultural landscapes.
At the moment, this implementation is still in development.

**Sources:**
    - Bauer, H.-G., Bezzel, E., & Fiedler, W. (Eds.). (2012). Das Kompendium
      der Vögel Mitteleuropas: Ein umfassendes Handbuch zu Biologie, Gefährdung
      und Schutz (Einbändige Sonderausg. der 2., vollständig überarb. und erw.
      Aufl. 2005). AULA-Verlag
    - Delius, J. D. (1965). A Population Study of Skylarks Alauda Arvensis.
      Ibis, 107(4), 466–492. https://doi.org/10.1111/j.1474-919X.1965.tb07332.x
    - Donald et al. (2002). Survival rates, causes of failure and productivity
      of Skylark Alauda arvensis nests on lowland farmland. Ibis, 144(4), 652–664.
      https://doi.org/10.1046/j.1474-919X.2002.00101.x
    - Glutz von Blotzheim, Urs N. (Ed.). (1985). Handbuch der Vögel Mitteleuropas.
      Bd. 10. Passeriformes (Teil 1) 1. Alaudidae - Hirundidae. AULA-Verlag, Wiesbaden.
      ISBN 3-89104-019-9
    - Jenny, M. (1990). Territorialität und Brutbiologie der Feldlerche Alauda
      arvensis in einer intensiv genutzten Agrarlandschaft. Journal für Ornithologie,
      131(3), 241–265. https://doi.org/10.1007/BF01640998
    - Püttmanns et al. (2022). Habitat use and foraging parameters of breeding Skylarks
      indicate no seasonal decrease in food availability in heterogeneous farmland.
      Ecology and Evolution, 12(1), e8461. https://doi.org/10.1002/ece3.8461
"""
@species Skylark begin
    # species parameters
    const movementrange::Length = 500m #XXX arbitrary
    const visionrange::Length = 200m #XXX arbitrary
    
    const eggtime::Int64 = 11 # days from laying to hatching
    const nestlingtime::Int64 = 9 # days from hatching to leaving nest
    const fledglingtime::Int64 = 21 # days from leaving the nest to independence

    #XXX predation mortality should be habitat-dependent
    const eggpredationmortality::Float64 = 0.03 # per-day egg mortality from predation
    const nestlingpredationmortality::Float64 = 0.03 # per-day nestling mortality from predation
    const fledglingpredationmortality::Float64 = 0.01 # per-day fledgling mortality from predation
    const firstyearmortality::Float64 = 0.38 # total mortality in the first year after independence
    const migrationmortality::Float64 = 0.33 # chance of dying during the winter

    const minimumterritory = 5000m² # size of territory under ideal conditions
    const mindistancetoedge = 60m # minimum distance of habitat to vertical structures
    const maxforageheight = 50cm # maximum preferred vegetation height for foraging
    const maxforagecover = 0.7 # maximum preferred vegetation cover for foraging
    const nestingheight = (15cm, 25cm) # min and max preferred vegetation height for nesting
    const nestingcover = (0.2, 0.5) # min and max preferred vegetation cover for nesting

    const matefaithfulness = 0.5 # chance of a female retaining her previous partner
    const nestingbegin::Tuple{Int64,Int64} = (April, 10) # begin nesting in the middle of April
    const nestbuildingtime::UnitRange{Int64} = 4:5 # 4-5 days needed to build a nest (doubled for first nest)
    const eggsperclutch::UnitRange{Int64} = 2:5 # eggs laid per clutch
    const nestingend::Int64 = July # last month of nesting

    # individual variables
    timer::Int64 = 0 # a counter that can be used for different purposes
    migrationdates::Tuple = () # is defined by each individual in @create(Skylark)
    mate::Int64 = -1 # the agent ID of the mate (-1 if none)
    nest::Tuple = () # coordinates of current nest
    clutch::Int64 = 0 # number and life stage of offspring in current clutch
end


## LIFE PHASES

#TODO respect habitat when moving

"""
Non-breeding adults move around with other individuals and check for migration.
"""
@phase Skylark nonbreeding begin
    # flocking behaviour - follow a random neighbour or move randomly
    neighbours = @neighbours(self.visionrange) #XXX check for the closest neighbour(s)?
    isempty(neighbours) ?
        @walk("random", self.movementrange) :
        @follow(@rand(neighbours), 30m) #XXX magic number
    # check if the bird migrates
    leave, arrive = self.migrationdates
    if leave <= model.date < arrive
        @kill(self.migrationmortality, "migration")
        self.sex == male ?
            @setphase(territorysearch) :
            @setphase(matesearch)
        @migrate(arrive)
        self.migrationdates = self.migrationdates .+ Year(1)
    end
end

"""
Males returning from migration move around to look for suitable habitats to establish a territory.
"""
@phase Skylark territorysearch begin
    if !isempty(self.territory)
        @setphase(occupation) # If the male still has a territory, it is occupied again
    elseif month(model.date) > self.nestingend
        @setphase(nonbreeding) # If the breeding season is over, give up looking
    else # otherwise check whether this is a suitable location
        newterritory = findterritory(self, model)
        if isempty(newterritory)
            @walk("random", self.movementrange)
        else
            for p in newterritory
                @occupy(p)
            end
            @debug("$(animalid(self)) has found a territory.")
        end
    end
end

"""
Once a male has found a territory, he remains in it until the breeding season is over,
adjusting it to new conditions when and as necessary.
"""
@phase Skylark occupation begin
    #move to a random location in the territory
    @move(@rand(self.territory))
    if month(model.date) > self.nestingend #XXX check that all the young have left the nest
        @setphase(nonbreeding)
    end
    #TODO adjust territory as needed (e.g. once a week, or when a brood is done?)
end

"""
Females returning from migration move around to look for a suitable partner with a territory.
"""
@phase Skylark matesearch begin
    if self.mate != -1 # if the female already had a partner last year...
        if isalive(@animal(self.mate))
            if @chance(self.matefaithfulness)
                # ...check if he is still alive and the female wants to stay with him
                #XXX is mate-faithfulness decided by the female when she returns,
                # or by the male taking whichever female comes to him first?
                @debug "$(animalid(self)) and $(animalid(n)) have mated again."
                @setphase(nesting)
                return
            else
                self.mate = -1
                @animal(self.mate).mate = -1
            end
        else
            self.mate = -1
        end
    end
    # look for a mate among the neighbouring birds, or move randomly
    for n in @neighbours(self.visionrange)
        if n.sex == male && n.mate == -1 && n.phase == occupation
            self.mate = n.id  #XXX may be the old partner again by chance
            n.mate = self.id
            @setphase(nesting)
            @debug "$(animalid(self)) and $(animalid(n)) have mated."
            return
        end
    end
    #@debug("$(animalid(self)) didn't find a mate.")
    if month(model.date) > self.nestingend # stop trying to find a mate if it's too late
        @setphase(nonbreeding)
    else
        @walk("random", self.movementrange)
    end
end

"""
Females that have found a partner build a nest and lay eggs in a suitable location.
"""
@phase Skylark nesting begin
    if model.date < Date(year(model.date), self.nestingbegin...)
        # wait for nesting to begin, moving around in the territory
        @move(@rand(@animal(self.mate).territory))
    elseif month(model.date) > self.nestingend
        @setphase(nonbreeding) # stop trying to nest if it's too late (should be rare)
    elseif isempty(self.nest)
        # choose site, build nest & lay eggs
        for pos in @shuffle!(deepcopy(@animal(self.mate).territory))
            if allowsnesting(self, model, pos)
                @move(pos)
                self.nest = pos
                self.clutch = @rand(self.eggsperclutch)
                # time to build + 1 day per egg laid
                self.timer = @rand(self.nestbuildingtime) + self.clutch
                if month(model.date) == self.nestingbegin[1] 
                    # the first nest takes twice as long to build
                    #XXX this may affect the first two nests
                    self.timer += self.timer*2-self.clutch
                end
                break
            end
        end
        isempty(self.nest) && @warn("$(animalid(self)) didn't find a nesting location.")
    elseif self.timer == 0
        @debug("$(animalid(self)) has laid $(self.clutch) eggs.")
        @setphase(breeding)
    else
        self.timer -= 1
    end
    # tillage and harvest destroys the nest
    @respond(tillage, destroynest!(self, "tillage"))
    @respond(harvesting, destroynest!(self, "harvesting"))
end

"""
Females that have laid eggs take care of their chicks, restarting the nesting process once the
chicks are independent or in case of brood loss.
"""
@phase Skylark breeding begin
    #XXX Schachtelbruten - sometimes skylarks start a new nest before the previous young are gone
    # wait for eggs to hatch & chicks to mature, checking for predation and disturbance mortality
    self.timer += 1
    #XXX this should be habitat-dependent!
    if self.timer <= self.eggtime
        @chance(self.eggpredationmortality) && destroynest!(self, "predation")
    elseif self.timer <= self.eggtime + self.nestlingtime
        @chance(self.nestlingpredationmortality) && destroynest!(self, "predation")
    elseif self.timer <= self.eggtime + self.nestlingtime + self.fledglingtime
        @chance(self.fledglingpredationmortality) && destroynest!(self, "predation")
    else
        # create new young, reset timer and clutch counter
        @reproduce(self.clutch, self.mate)
        self.clutch = 0
    end
    if self.clutch > 0
        # tillage and harvest destroys the nest
        @respond(tillage, destroynest!(self, "tillage"))
        @respond(harvesting, destroynest!(self, "harvesting"))        
    else # restart breeding cycle if there is time
        self.timer = 0
        if month(model.date) <= self.nestingend
            @setphase(nesting)
        elseif month(model.date) > self.nestingend
            @setphase(nonbreeding)
        end
    end
end


## SUPPORTING FUNCTIONS

"""
    migrationdates(skylark, model)

Select the dates on which this skylark will leave for / return from its migration,
based on observed migration patterns.
"""
function migrationdates(skylark::Skylark, model::SimulationModel)
    #TODO this ought to be temperature-dependent and dynamic
    minleave = skylark.sex == female ? (September, 15) : (October, 1)
    minarrive = skylark.sex == male ? (February, 15) : (March, 1)
    deltaleave = @rand(0:45) #XXX ought to be normally distributed
    deltaarrive = @rand(0:15) #XXX ought to be normally distributed
    leave = Date(year(model.date), minleave[1], minleave[2]) + Day(deltaleave)
    arrive = Date(year(model.date)+1, minarrive[1], minarrive[2]) + Day(deltaarrive)
    (leave, arrive)
end

"""
    findterritory(skylark, model)

Check whether the habitat surrounding the skylark is suitable for establishing a territory.
If it is, return the list of coordinates that make up the new territory, else return an empty list.
"""
function findterritory(skylark::Skylark, model::SimulationModel)
    effectivesize::Area = 0m² # the usable size of the territory, weighted by habitat quality
    territory::Vector{Tuple{Int64,Int64}} = []
    msize = size(model.landscape)
    radius = 0
    constrained = false
    # Inspect the landscape in concentric circles around the individual until enough pixels have
    # been found to provide a territory of sufficient size and quality. If there are no suitable
    # pixels in one circle, break off the search (territories must be contiguous).
    while !constrained && effectivesize < skylark.minimumterritory
        constrained = true
        if radius == 0
            coords = [skylark.pos]
        else # list all coordinates in the next circle...
            coords = []
            xrange = (skylark.pos[1]-radius, skylark.pos[1]+radius)
            yrange = (skylark.pos[2]-radius, skylark.pos[2]+radius)
            for x in xrange[1]:xrange[2]
                push!(coords, (x, yrange[1]))
                push!(coords, (x, yrange[2]))
            end
            for y in (yrange[1]+1):(yrange[2]-1) #avoid duplicating the corners
                push!(coords, (y, xrange[1]))
                push!(coords, (y, xrange[2]))
            end
        end
        #FIXME some duplicates remain?
        for c in coords # ...then inspect them
            (c[1] <= 0 || c[2] <= 0 || c[1] > msize[1] || c[2] > msize[2]) && continue
            (isoccupied(model, "Skylark", c)) && continue
            push!(territory, c)
            quality = foragequality(skylark, model, c)
            effectivesize += @param(world.mapresolution)^2*quality
            (quality > 0) && (constrained = false)
            #XXX check for nesting habitats?
        end
        radius +=1
    end
    constrained ? [] : territory
end

"""
    foragequality(skylark, model, pos)

Calculate the relative quality of the habitat at this position for foraging.
This assumes that open habitat is best (quality = 1.0), and steadily decreases as vegetation
height and/or cover increase. (Linear regressions based on Püttmanns et al., 2021; Jeromin, 2002;
Jenny, 1990b.)
"""
function foragequality(skylark::Skylark, model::SimulationModel, pos::Tuple{Int64,Int64})
    #TODO this is a key function that needs to be validated thoroughly
    if !(@landcover() in (grass, soil, agriculture)) ||
        (@distanceto(forest) < skylark.mindistancetoedge) ||
        (@distanceto(builtup) < skylark.mindistancetoedge)
        return 0.0
    end
    groundcoverfactor = x -> bounds((-1/skylark.maxforagecover)*x + 1.0, max=1.0)
    plantheightfactor = x -> bounds((-1/skylark.maxforageheight)*(x |> cm) + 1.0, max=1.0)
    return bounds(groundcoverfactor(@cropcover()) + plantheightfactor(@cropheight()), max=1.0)
end

"""
    allowsnesting(skylark, model, pos)

Check whether the given position is suitable for nesting.
"""
function allowsnesting(skylark::Skylark, model::SimulationModel, pos::Tuple{Int64,Int64})
    #TODO is this condition correct? -> needs validation!
    (@landcover() == grass ||
     (@landcover() == agriculture &&
      (skylark.nestingheight[1] <= (@cropheight()) <= skylark.nestingheight[2]) &&
      (skylark.nestingcover[1] <= @cropcover() <= skylark.nestingcover[2]))) #&&
    #FIXME if we add the distance requirement, females don't find a nesting spot?
      #(@distanceto(forest) < skylark.mindistancetoedge) &&
      #(@distanceto(builtup) < skylark.mindistancetoedge)
end

"""
    destroynest!(skylark, reason)

Remove the skylark's nest and offspring due to disturbance or predation.
"""
function destroynest!(self::Skylark, reason::String)
    self.nest = ()
    self.clutch = 0
    @debug("$(animalid(self)) had her nest destroyed by $reason.")
end

## INITIALISATION

"""
Initialise a skylark individual. Selects migration dates and checks if the bird
should currently be on migration. Also sets other individual-specific variables.
"""
@create Skylark begin
    @debug "Added $(animalid(self)) at $(self.pos)"
    self.migrationdates = migrationdates(self, model) .- Year(1) #XXX should be dynamic
    leave, arrive = self.migrationdates
    if leave <= model.date < arrive
        self.sex == male ?
            @setphase(territorysearch) :
            @setphase(matesearch)
        @migrate(arrive)
        self.migrationdates = self.migrationdates .+ Year(1)
    end
end

@populate Skylark begin
    # initialise on open land, at least 60m from vertical structures
    habitat = @habitat((@landcover() == grass || @landcover() == agriculture) &&
                       @distanceto(forest) >= 60m && @distanceto(builtup) >=60m) #XXX magic numbers
    initphase = nonbreeding
    birthphase = nonbreeding
    indarea = 3ha #XXX magic number
    pairs = true
end
